#pragma inline
//---------------------------------------------------------------------------
#include <math.h>
#include "AJBLisp.h"
#include "random.h"

//---------------------------------------------------------------------------
//
// NOTE: Using dotted lists as arguments to any control structure will
//       crash the interpreter.
//       This is not a bug but a concious decision to avoid numerous
//       tests for badly formed lists.
//       There is no valid interpretation of such forms in any case.
//       e.g. (cond ((eq x 3) (dothis) . (dothat)) (t . nil)) = BOOM!

LISP Endloop;          // LOOP control variable
char Filename[241];    // Filename for LOAD

// (QUOTE <literal>) or '<literal>
LISP Quote(LISP arg)
{
  if(arg->cdr)
    LispError(2);   // Incorrect number of args
  return arg->car;
}

// (EQ <this> <that>)
// Identical atoms or same list or string
LISP Eq(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  return _Eq(args[0], args[1]);
}

// Helper function for Eq
LISP _Eq(LISP a, LISP b)
{
  if(a == b) return T;
  return NIL;
}

// (EQL <this> <that>)
// Identical symbols or lists, equal numbers or strings
LISP Eql(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  return _Eql(args[0], args[1]);
}

// Helper function for Eql
LISP _Eql(LISP a, LISP b)
{
  if(!a && !b) return T;
  if(!a || !b) return NIL;
  if(a->flags != b->flags) return NIL;
  if(IsInteger(a))
    return (a->ivalue == b->ivalue) ? T : NIL;
  if(IsReal(a))
    return (a->rvalue == b->rvalue) ? T : NIL;
  if(IsString(a))
    return (!strcmp(a->svalue->text, b->svalue->text)) ? T : NIL;
  return (a == b) ? T : NIL;
}

// (EQUAL <this> <that>)
// Identical atoms, equal numbers, strings with identical text
// or equivalent lists
LISP Equal(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  return _Equal(args[0], args[1]);
}

// helper function for Equal
LISP _Equal(LISP a, LISP b)
{
  if(_Eql(a, b)) return T;
  if(IsAtom(a) || IsAtom(b)) return NIL;
  return(_Equal(a->car, b->car) && _Equal(a->cdr, b->cdr)) ? T : NIL;
}

// (CONS <car> <cdr>)
LISP Cons(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  return NewCons(args[0], args[1]);
}

// (NOT <boolean>)
LISP Not(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  return (args[0]) ? NIL : T;
}

// (AND [<boolean> ...])
LISP And(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);

  while(pargs)
  {
    if(!Eval(pargs->car))
    {
      PPLIST = ppsave;
      return NIL;
    }
    pargs = pargs->cdr;
  }
  PPLIST = ppsave;
  return T;
}

// (OR [<boolean> ...])
LISP Or(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);

  while(pargs)
  {
    if(Eval(pargs->car))
    {
      PPLIST = ppsave;
      return T;
    }
    pargs = pargs->cdr;
  }
  PPLIST = ppsave;
  return NIL;
}

// (BNOT <integer>)
LISP Bnot(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(!IsInteger(args[0]))
    LispError(5); // Integer argument expected
  return NewInteger(~args[0]->ivalue);

}

// (BAND <integer>)
LISP Band(int nargs, LISP args[])
{
  int result = ~0;
  for(int i = 0; i < nargs; i++)
  {
    if(!IsInteger(args[i]))
      LispError(5); // Integer argument expected
    result &= args[i]->ivalue;
  }
  return NewInteger(result);
}

// (BOR <integer>)
LISP Bor(int nargs, LISP args[])
{
  int result = 0;
  for(int i = 0; i < nargs; i++)
  {
    if(!IsInteger(args[i]))
      LispError(5); // Integer argument expected
    result |= args[i]->ivalue;
  }
  return NewInteger(result);
}

// (SET <symbol> <value>)
LISP Set(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  return _Set(args[0], args[1]);
}

// Helper function for Set
LISP _Set(LISP c, LISP value)
{
  if(!IsSymbol(c))
    LispError(7); // Symbol argument expected
  if(c <= QUOTE)
    LispError(13); // Cannot change value of special symbol
  c->value = value;
  return(value);
}

// (SETQ <literal> <value>)
LISP Setq(LISP arg)
{
  PP* ppsave = PPLIST;
  volatile LISP lit = NIL;
  PP plit;
  protect(&lit, &plit);
  if(IsAtom(arg))
    LispError(2); // Incorrect number of arguments
  lit = arg->car;
  LISP val = arg->cdr;
  if(IsAtom(val) || val->cdr)
    LispError(2); // Incorrect number of arguments
  val = Eval(val->car);
  PPLIST = ppsave;
  return _Set(lit, val);
}

// (CAR <list>)
LISP Car(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]))
    LispError(3); // List argument expected
  return (args[0])->car;
}

// (CDR <list>)
LISP Cdr(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]))
    LispError(3); // List argument expected
  return (args[0])->cdr;
}

// (CAAR <list>)
LISP Caar(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]) || IsAtom(args[0]->car))
    LispError(3); // List argument expected
  return (*args)->car->car;
}

// (CADR <list>)
LISP Cadr(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]) || IsAtom(args[0]->cdr))
    LispError(3); // List argument expected
  return (*args)->cdr->car;
}

// (CDAR <list>)
LISP Cdar(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]) || IsAtom(args[0]->car))
    LispError(3); // List argument expected
  return (*args)->car->cdr;
}

// (CDDR <list>)
LISP Cddr(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]) || IsAtom(args[0]->cdr))
    LispError(3); // List argument expected
  return (*args)->cdr->cdr;
}

// (RPLACA <list> <value>)
LISP Rplaca(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]))
    LispError(3); // List argument expected
  args[0]->car = args[1];
  return args[0];
}

// (RPLACD <list> <value>)
LISP Rplacd(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(IsAtom(args[0]))
    LispError(3); // List argument expected
  args[0]->cdr = args[1];
  return args[0];
}

// (EVAL <list>)
LISP _Eval(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  return Eval(args[0]);
}

// (LENGTH <list>)
LISP Length(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(!args[0])
    return NewInteger(0);
  if(IsAtom(args[0]))
    LispError(3); // List argument expected
  int i = 0;
  LISP c = args[0];
  while(c)
  {
    c = c->cdr;
    i++;
  }
  return NewInteger(i);
}

// (+ [<num> ...])
LISP Add(int nargs, LISP args[])
{
  int result = 0;
  double fresult = 0.0;
  bool intf = true;

  while(nargs)
  {
    if(!IsNumber(*args))
      LispError(4); // Numeric argument expected
    if(intf)
    {
      if(IsReal(*args))
      {
        intf = false;
        fresult = result;
        continue;
      };
      result += (*args)->ivalue;
      asm into;              // Overflow test
      args++;
      nargs--;
    }
    else
    {
      if(IsReal(*args))
        fresult += (*args++)->rvalue;
      else
        fresult += (*args++)->ivalue;
      nargs--;
    }
  }
  if(intf)
    return NewInteger(result);
  else
    return NewReal(fresult);
}

// (- <num> [<num>])
LISP Subtract(int nargs, LISP args[])
{
  if(nargs == 1)    // Negate
  {
    if(!IsNumber(args[0]))
      LispError(4); // Numeric argument expected
    if(IsInteger(args[0]))
      return NewInteger(-args[0]->ivalue);
    else
      return NewReal(-args[0]->rvalue);
  }
  if(nargs == 2)   // Subtract
  {
    if(!IsNumber(args[0]) || !IsNumber(args[1]))
      LispError(4); // Numeric argument expected
    if(IsReal(args[0]))
    {
      if(IsReal(args[1]))
        return NewReal(args[0]->rvalue - args[1]->rvalue);
      else
        return NewReal(args[0]->rvalue - args[1]->ivalue);
    }
    else
    {
      if(IsReal(args[1]))
        return NewReal(args[0]->ivalue - args[1]->rvalue);
      else
      {
        int i = args[0]->ivalue - args[1]->ivalue;
        asm into;              // Overflow test
        return NewInteger(i);
      }
    }
  }
  LispError(2);   // Incorrect number of args
  return(NIL);  // Keep compiler happy
}

// (* [<num> ...])
LISP Multiply(int nargs, LISP args[])
{
  int result = 1;
  double fresult = 1.0;
  bool intf = true;

  while(nargs)
  {
    if(!IsNumber(*args))
      LispError(4); // Numeric argument expected
    if(intf)
    {
      if(IsReal(*args))
      {
        intf = false;
        fresult = result;
        continue;
      };
      result *= (*args)->ivalue;
      asm into;              // Overflow test
      args++;
      nargs--;
    }
    else
    {
      if(IsReal(*args))
        fresult *= (*args++)->rvalue;
      else
        fresult *= (*args++)->ivalue;
      nargs--;
    }
  }
  if(intf)
    return NewInteger(result);
  else
    return NewReal(fresult);
}

// (/ <num> <num>)
LISP Divide(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]) || !IsNumber(args[1]))
    LispError(4); // Numeric argument expected
  if(IsReal(args[1]))
  {
    if(IsReal(args[0]))
      return NewReal(args[0]->rvalue / args[1]->rvalue);
    else
      return NewReal(args[0]->ivalue / args[1]->rvalue);
  }
  else
  {
    if(IsReal(args[0]))
      return NewReal(args[0]->rvalue / args[1]->ivalue);
    else
      return NewInteger(args[0]->ivalue / args[1]->ivalue);
  }
}

// (REM <integer> <integer>)
LISP Rem(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(!IsInteger(args[0]) || !IsInteger(args[1]))
    LispError(5); // Integer argument expected
  return NewInteger(args[0]->ivalue % args[1]->ivalue);
}

// (> <num> <num>)
LISP Greaterp(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]) || !IsNumber(args[1]))
    LispError(4); // Numeric argument expected
  if(IsReal(args[0]))
  {
    if(IsReal(args[1]))
      return (args[0]->rvalue > args[1]->rvalue) ? T : NIL;
    else
      return (args[0]->rvalue > args[1]->ivalue) ? T : NIL;
  }
  else
  {
    if(IsReal(args[1]))
      return (args[0]->ivalue > args[1]->rvalue) ? T : NIL;
    else
      return (args[0]->ivalue > args[1]->ivalue) ? T : NIL;
  }
}

// (< <num> <num>)
LISP Lessp(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]) || !IsNumber(args[1]))
    LispError(4); // Numeric argument expected
  if(IsReal(args[0]))
  {
    if(IsReal(args[1]))
      return (args[0]->rvalue < args[1]->rvalue) ? T : NIL;
    else
      return (args[0]->rvalue < args[1]->ivalue) ? T : NIL;
  }
  else
  {
    if(IsReal(args[1]))
      return (args[0]->ivalue < args[1]->rvalue) ? T : NIL;
    else
      return (args[0]->ivalue < args[1]->ivalue) ? T : NIL;
  }
}

// (<= <num> <num>)
LISP NGreaterp(int nargs, LISP args[])
{
  return Lessp(nargs, args) ? NIL : T;
}

// (>= <num> <num>)
LISP NLessp(int nargs, LISP args[])
{
  return Lessp(nargs, args) ? NIL : T;
}

// (= <num> <num>)
LISP Equalp(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]) || !IsNumber(args[1]))
    LispError(4); // Numeric argument expected
  if(IsReal(args[0]))
  {
    if(IsReal(args[1]))
      return (args[0]->rvalue == args[1]->rvalue) ? T : NIL;
    else
      return (args[0]->rvalue == args[1]->ivalue) ? T : NIL;
  }
  else
  {
    if(IsReal(args[1]))
      return (args[0]->ivalue == args[1]->rvalue) ? T : NIL;
    else
      return (args[0]->ivalue == args[1]->ivalue) ? T : NIL;
  }
}

// (<> <num> <num>)
LISP NEqualp(int nargs, LISP args[])
{
  return Equalp(nargs, args) ? NIL : T;
}

// (ZEROP <num>)
LISP Zerop(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsInteger(args[0]) && (args[0]->ivalue == 0)) return T;
  if(IsReal(args[0]) && (args[0]->rvalue == 0.0)) return T;
  return NIL;
}

// (ONEP <num>)
LISP Onep(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsInteger(args[0]) && (args[0]->ivalue == 1)) return T;
  if(IsReal(args[0]) && (args[0]->rvalue == 1.0)) return T;
  return NIL;
}

// (MINUSP <num>)
LISP Minusp(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsInteger(args[0]) && (args[0]->ivalue < 0)) return T;
  if(IsReal(args[0]) && (args[0]->rvalue < 0.0)) return T;
  return NIL;
}

// (ATOM <arg>)
LISP Atom(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsAtom(args[0])) return T;
  else return NIL;
}

// (SYMBOLP <arg>)
LISP Symbolp(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsSymbol(args[0])) return T;
  else return NIL;
}

// (NUMBERP <arg>)
LISP Numberp(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsNumber(args[0])) return T;
  else return NIL;
}

// (LISTP <arg>)
LISP Listp(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsAtom(args[0])) return NIL;
  else return T;
}

// (STRINGP <arg>)
LISP Stringp(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsString(args[0])) return T;
  else return NIL;
}

// (SUBRP <arg>)
LISP Subrp(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsSubr(args[0])) return T;
  else return NIL;
}

// (FSUBRP <arg>)
LISP FSubrp(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(IsFSubr(args[0])) return T;
  else return NIL;
}

// (IF <cond> <then> [<else>])
LISP If(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP _then = NIL;
  PP p_then;
  protect(&_then, &p_then);
  volatile LISP _else = NIL;
  PP p_else;
  protect(&_else, &p_else);
  // Extract condition
  LISP _if = args->car;
  args = args->cdr;
  if(IsAtom(args))
    LispError(2); // Incorrect number of arguments to function
  // Extract clauses
  _then = args->car;
  _else = args->cdr;
  if(_else) // Else clause may be null
  {
    // If present _else must be a singleton list
    if(IsAtom(_else))
      LispError(3); // List argument expected
    if(_else->cdr)
      LispError(2); // Incorrect number of arguments to function
    // Set else clause to list item
    _else = _else->car;
  }
  LISP result = (Eval(_if)) ? Eval(_then) : Eval(_else);
  PPLIST = ppsave;
  return result;
}

// (COND (<cond>  <action> ...) ...)
LISP Cond(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  volatile LISP clause = NIL;
  PP pclause;
  protect(&clause, &pclause);
  volatile LISP action = NIL;
  PP paction;
  protect(&action, &paction);

  while(pargs)              // While there are clauses left to test
  {
    LISP clause = pargs->car;
    if(IsAtom(clause))
      LispError(3);   // List argument expected
    LISP test = Eval(clause->car);
    // test is condition part of clause, it is unprotected because
    // there are no heap operations between initialisation and use
    if(test)
    {
      action = clause->cdr;
      if(!action)          // No action part, return value of test part
      {
        PPLIST = ppsave;
        return test;
      }
      while(true)           // Implicit progn in action part
      {
        LISP result = Eval(action->car);
        if(!action->cdr)
        {
          PPLIST = ppsave;
          return result;
        }
        action = action->cdr;
      }
    }
    pargs = pargs->cdr;     // Next clause
  }
  PPLIST = ppsave;          // No more clauses, return NIL
  return NIL;
}

// (LOOP <body> ...)
LISP Loop(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  volatile LISP start = args;
  PP pstart;
  protect(&start, &pstart);
  Push(Endloop);
  Endloop = NIL;
  if(!start)
    LispError(16); // Aborted null LOOP
  LISP result;
  while(!Endloop)
  {
    if(!pargs) pargs = start;   // Loop
    result = Eval(pargs->car);
    pargs = pargs->cdr;
  }
  Endloop = Pop();
  PPLIST = ppsave;
  return result;
}

// (UNTIL <cond> <body> ...)
LISP Until(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  LISP result = Eval(pargs->car);
  if(!result)
  {
    PPLIST = ppsave;
    return result;
  }
  Endloop = T;
  while(true)
  {
    pargs = pargs->cdr;
    if(!pargs)
    {
      PPLIST = ppsave;
      return result;
    }
    result = Eval(pargs->car);
  }
}

// (WHILE <cond> <body> ...)
LISP While(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  LISP result = Eval(pargs->car);
  if(result)
  {
    PPLIST = ppsave;
    return result;
  }
  Endloop = T;
  while(true)
  {
    pargs = pargs->cdr;
    if(!pargs)
    {
      PPLIST = ppsave;
      return result;
    }
    result = Eval(pargs->car);
  }
}

// (PROGN <body> ...)
LISP Progn(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  LISP result = NIL;
  while(true)
  {
    if(!pargs)
    {
      PPLIST = ppsave;
      return result;
    }
    result = Eval(pargs->car);
    pargs = pargs->cdr;
  }
}

// (LET ((<symbol> <value>) ...) <body>)
LISP Let(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  volatile LISP bindings = args->car;
  PP pbindings;
  protect(&bindings, &pbindings);
  NewFrame(NIL);
  base->save = true;
  // Evaluate the bindings into the stack frame
  while(bindings)
  {
    if(!IsCons(bindings->car) || !IsSymbol(bindings->car->car))
      LispError(24);            // Badly formed binding in LET
    Push(NIL);                  // Space for symbol pointer
    LISP val = bindings->car->cdr;
    if(val)
      val = Eval(val->car);
    Push(val);
    base->nargs++;
    bindings = bindings->cdr;
  }
  // Create the bindings
  bindings = args->car;
  LISP* locals = base->args;
  while(bindings)
  {
    LISP variable = bindings->car->car;
    *locals++ = variable;
    LISP oldval = variable->value;
    variable->value = *locals;
    *locals++ = oldval;
    bindings = bindings->cdr;
  }
  // Evaluate the body
  LISP result = NIL;
  pargs = pargs->cdr;
  while(true)
  {
    if(!pargs)
      break;
    result = Eval(pargs->car);
    pargs = pargs->cdr;
  }
  PPLIST = ppsave;
  PopFrame();
  return result;
}

// (SLET ((<symbol> <value>) ...) <body>)
LISP Slet(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  volatile LISP bindings = args->car;
  PP pbindings;
  protect(&bindings, &pbindings);
  NewFrame(NIL);
  base->save = true;
  // Evaluate the bindings into the stack frame
  while(bindings)
  {
    if(!IsCons(bindings->car) || !IsSymbol(bindings->car->car))
      LispError(24);            // Badly formed binding in LET
    LISP val = bindings->car->cdr;
    if(val)
      val = Eval(val->car);
    LISP variable = bindings->car->car;
    Push(variable);
    LISP oldval = variable->value;
    variable->value = val;
    Push(oldval);
    base->nargs++;
    bindings = bindings->cdr;
  }
  // Evaluate the body
  LISP result = NIL;
  pargs = pargs->cdr;
  while(true)
  {
    if(!pargs)
      break;
    result = Eval(pargs->car);
    pargs = pargs->cdr;
  }
  PPLIST = ppsave;
  PopFrame();
  return result;
}

// (LIST [<arg> ...])
LISP List(LISP args)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  volatile LISP result = NewCons(NIL, NIL);
  PP presult;
  protect(&result, &presult);
  volatile LISP tail = result;
  PP ptail;
  protect(&tail, &ptail);
  while(pargs)
  {
    tail->cdr = NewCons(Eval(pargs->car), NIL);
    tail = tail->cdr;
    pargs = pargs->cdr;
  }
  PPLIST = ppsave;
  return result->cdr;
}

// (MEMBER [<arg> ...])
LISP Member(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsAtom(args[0]) || IsAtom(args[1]))
    LispError(1); // Bad Arguments
  return _Member(args[0], args[1]);
}

// Helper function for Member
LISP _Member(LISP a, LISP b)
{
  if(!b) return NIL;
  if(_Eql(a, b->car)) return T;
  return _Member(a, b->cdr);
}

// (OPEN <filename> <mode>)
// This function is the only way to create a File atom
LISP Open(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]) || !IsString(args[1]))
    LispError(1); // Bad Arguments
  FILE* f = fopen(args[0]->svalue->text, args[1]->svalue->text);
  if(!f) return NIL;
  // Use NewInteger to make the File cell
  LISP result = NewInteger((int)f);
  // Then correctly set the type
  result->flags = tFILE;
  return result;
}

// (CLOSE <file>)
LISP Close(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!args[0]) return NIL;    // It wasn't opened
  if(!IsFile(args[0]))
    LispError(1); // Bad Arguments
  int result = fclose(args[0]->file);
  return result ? NIL : T;
}

// (FLOAT-FORMAT <string>)
LISP FloatFormat(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]))
    LispError(8);   // String argument expected
  strncpy(&ffmt[1], args[0]->svalue->text, 9);
  ffmt[0] = '%';
  ffmt[10] = '\0';
  return args[0];
}

// (INTEGER-FORMAT <string>)
LISP IntegerFormat(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]))
    LispError(8);   // String argument expected
  strncpy(&ifmt[1], args[0]->svalue->text, 8);
  ifmt[0] = '%';
  ifmt[10] = '\0';
  return args[0];
}

// Function to handle all output
LISP _print(LISP args, bool cflag, bool cr)
{
  PP* ppsave = PPLIST;
  volatile LISP pargs = args;
  PP ppargs;
  protect(&pargs, &ppargs);
  LISP result = NIL;
  FILE* out = stdout;
  LISP f;
  if(pargs && IsFile(f = Eval(pargs->car)))
  {
    out = f->file;
    pargs = pargs->cdr;
  }
  while(pargs)
  {
    print(result = Eval(pargs->car), out, cflag);
    pargs = pargs->cdr;
  }
  if(cr) fputc('\n', out);
  PPLIST = ppsave;
  return result;
}

// (PRIN [<file>] [<arg> ...])
LISP Prin(LISP args)
{
  return _print(args, false, false);
}

// (PRINC [<file>] [<arg> ...])
LISP Princ(LISP args)
{
  return _print(args, true, false);
}

// (PRINT [<file>] [<arg> ...])
LISP Print(LISP args)
{
  return _print(args, false, true);
}

// (PRINTC [<file>] [<arg> ...])
LISP Printc(LISP args)
{
  return _print(args, true, true);
}

// (READ [<file>])
LISP Read(int nargs, LISP args[])
{
  if(nargs == 0)
    return read(stdin);
  if((nargs == 1) && IsFile(args[0]))
    return read(args[0]->file);
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  LispError(1); // Bad Arguments
  return NIL;
}

// (READLINE [<file>])
LISP Readline(int nargs, LISP args[])
{
  if(nargs == 0)
    return readline(stdin);
  if((nargs == 1) && IsFile(args[0]))
    return readline(args[0]->file);
  LispError(9); // File argument expected
  return NIL;
}

// (FLUSH [<file>])
LISP Flush(int nargs, LISP args[])
{
  if(nargs == 0)
    return flush(stdin);
  if((nargs == 1) && IsFile(args[0]))
    return flush(args[0]->file);
  LispError(9); // File argument expected
  return NIL;
}

// (ABS <num>)
LISP Abs(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]))
    LispError(4); // Numeric argument expected
  return (IsInteger(args[0])) ? NewInteger(abs(args[0]->ivalue)) :
                                NewReal(fabs(args[0]->rvalue));
}

// Fetch a single numerical arg as a double
double MathsArgs(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]))
    LispError(4); // Numeric argument expected
  return (IsInteger(args[0])) ? (args[0]->ivalue) : (args[0]->rvalue);
}

// (FLOAT <num>)
LISP Float(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsString(args[0]))
  {
    double r;
    integer n;    // The number of characters converted
    int j = sscanf(args[0]->svalue->text, "%lf%n", &r, &n);
    if((j == 1) && (n == strlen(args[0]->svalue->text)))
      return(NewReal(r));
    else LispError(20); // Cannot interpret string as float
  }
  return NewReal(MathsArgs(nargs, args));
}

// (INTEGER <num>)
LISP Integer(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2); // Incorrect number of args
  if(IsString(args[0]))
  {
    int i;
    integer n;    // The number of characters converted
    int j = sscanf(args[0]->svalue->text, "%ld%n", &i, &n);
    if((j == 1) && (n == strlen(args[0]->svalue->text)))
      return(NewReal(i));
    else LispError(21); // Cannot interpret string as integer
  }
  double r = MathsArgs(nargs, args);
  return NewInteger((int)(r + ((r < 0) ? -0.5 : 0.5)));
}

// (FLOOR <num>)
LISP Floor(int nargs, LISP args[])
{
  double arg = MathsArgs(nargs, args);
  return NewInteger(floor(arg));
}

// (CEIL <num>)
LISP Ceil(int nargs, LISP args[])
{
  double arg = MathsArgs(nargs, args);
  return NewInteger(ceil(arg));
}

// (SQRT <num>)
LISP Sqrt(int nargs, LISP args[])
{
  double arg = MathsArgs(nargs, args);
  return NewReal(sqrt(arg));
}

// (SIN <num>)
LISP Sin(int nargs, LISP args[])
{
  return NewReal(sin(MathsArgs(nargs, args)));
}

// (COS <num>)
LISP Cos(int nargs, LISP args[])
{
  return NewReal(cos(MathsArgs(nargs, args)));
}

// (TAN <num>)
LISP Tan(int nargs, LISP args[])
{
  return NewReal(tan(MathsArgs(nargs, args)));
}

// (ASIN <num>)
LISP Asin(int nargs, LISP args[])
{
  double arg = MathsArgs(nargs, args);
  return NewReal(asin(arg));
}

// (ACOS <num>)
LISP Acos(int nargs, LISP args[])
{
  double arg = MathsArgs(nargs, args);
  return NewReal(acos(arg));
}

// (ATAN <num> [<num>])
LISP Atan(int nargs, LISP args[])
{
  if(nargs ==1)
  {
    if(!IsNumber(args[0]) )
      LispError(4); // Numeric argument expected
    if(IsInteger(args[0]))
      return NewReal(atan(args[0]->ivalue));
    else
      return NewReal(atan(args[0]->rvalue));
  }
  if(nargs != 2)
    LispError(2); // Incorrect number of args

  if(!IsNumber(args[0]) || !IsNumber(args[1]))
    LispError(4); // Numeric argument expected

  double x = (IsInteger(args[0])) ? args[0]->ivalue : args[0]->rvalue;
  double y = (IsInteger(args[1])) ? args[1]->ivalue : args[1]->rvalue;
  return NewReal(atan2(y, x));
}

// (LOG <num>)
LISP Log(int nargs, LISP args[])
{
  double arg = MathsArgs(nargs, args);
  return NewReal(log10(arg));
}

// (LN <num>)
LISP Ln(int nargs, LISP args[])
{
  double arg = MathsArgs(nargs, args);
  return NewReal(log(arg));
}

// (EXP <num>)
LISP Exp(int nargs, LISP args[])
{
  return NewReal(exp(MathsArgs(nargs, args)));
}

// (POWER <num> <num>)
LISP Power(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]) || !IsNumber(args[1]))
    LispError(4); // Numeric argument expected
  double arg1 = (IsReal(args[0])) ? args[0]->rvalue : args[0]->ivalue;
  double arg2 = (IsReal(args[1])) ? args[1]->rvalue : args[1]->ivalue;
  return NewReal(pow(arg1, arg2));
}

// (RANDOM)
LISP Random(int nargs, LISP args[])
{
  if(nargs == 0)
  return NewReal(ran1());
  if(nargs > 1)
    LispError(2); // Incorrect number of args
  if(!IsNumber(args[0]))
    LispError(4); // Numeric argument expected
  if(IsReal(args[0]))
    return NewReal(ran1() * args[0]->rvalue);
  else
    return NewInteger((int)floor(ran1() * args[0]->ivalue));
}

// (ASSOC <key> <list>)
LISP Assoc(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  return _Assoc(args[0], args[1]);
}

// Helper function for Assoc
LISP _Assoc(LISP x, LISP a)
{
  if(!IsSymbol(x))
    LispError(7); // Symbol argument expected
  if(!a) return NIL;
  while(a)
  {
    if(IsAtom(a))
      LispError(14); // Badly formed association list
    LISP e = a->car;
    if(IsAtom(e) || !IsSymbol(e->car))
      LispError(14); // Badly formed association list
    if(e->car == x) return e;
    a = a->cdr;
  }
  return NIL;
}

// (PUT <symbol> <key> <value>)
LISP Put(int nargs, LISP args[])
{
  if(nargs != 3)
    LispError(2);   // Incorrect number of args
  if(!IsSymbol(args[0]) || !IsSymbol(args[1]))
    LispError(1); // Bad Arguments
  LISP a = _Assoc(args[1], args[0]->prop);
  if(a)
    a->cdr = args[2];
  else
    args[0]->prop = NewCons(NewCons(args[1], args[2]), args[0]->prop);
  return args[2];
}

// (GET <symbol> <key>)
LISP Get(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsSymbol(args[0]) || !IsSymbol(args[1]))
    LispError(7); // Symbol argument expected
  LISP entry = _Assoc(args[1], args[0]->prop);
  if(entry) entry = entry->cdr;
  return entry;
}

// (REMPROP <symbol> <key>)
LISP Remprop(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsSymbol(args[0]) || !IsSymbol(args[1]))
    LispError(1); // Bad Arguments
  LISP* last = &(args[0]->prop);
  LISP a = *last;
  if(!a) return NIL;
  while(a)
  {
    if(IsAtom(a))
      LispError(15); // Badly formed property list
    LISP e = a->car;
    if(IsAtom(e) || !IsSymbol(e->car))
      LispError(15); // Badly formed property list
    if(e->car == args[1])
    {
      *last = a->cdr;
      return e->cdr;
    }
    last = &(a->cdr);
    a = *last;
  }
  return NIL;
}

// (PLIST <symbol>)
LISP Plist(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsSymbol(args[0]))
    LispError(7);   // Symbol argument expected
  return args[0]->prop;
}

// (OBLIST)
LISP Oblist(int nargs, LISP [])
{
  if(nargs != 0)
    LispError(2);   // Incorrect number of args
  PP* ppsave = PPLIST;
  volatile LISP next = NIL;
  PP pnext;
  protect(&next, &pnext);
  volatile LISP result = NewCons(NIL, NIL);
  PP presult;
  protect(&result, &presult);
  volatile LISP tail = result;
  PP ptail;
  protect(&tail, &ptail);
  for(int i = 0; i < HASHSIZE; i++)
  {
    next = hashtable[i];
    if(!next) continue;
    while(next)
    {
      if((next->value != UNDEFINED) || (next->prop != NIL))
      {
        tail->cdr = NewCons(next, NIL);
        tail = tail->cdr;
      }
      next = next->hash;
    }
  }
  PPLIST = ppsave;
  return result->cdr;
}

// (STRLEN <string>)
LISP Strlen(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]))
    LispError(1); // Bad Arguments
  return NewInteger(strlen(args[0]->svalue->text));
}

// (CONCAT <string> <string>)
LISP Concat(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]) || !IsString(args[0]))
    LispError(8); // String argument expected
  LISP result = NewString(strlen(args[0]->svalue->text) +
                          strlen(args[1]->svalue->text) + 1);
  strcpy(result->svalue->text, args[0]->svalue->text);
  strcat(result->svalue->text, args[1]->svalue->text);
  return result;
}

// (SUBSTRING <string> <int> [<int>])
LISP Substring(int nargs, LISP args[])
{
  int from;
  int size = -1;
  if((nargs < 2) || (nargs >3))
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]))
    LispError(8);   // String argument expected
  if(!IsInteger(args[1]))
    LispError(5);   // Integer argument expected
  from = args[1]->ivalue;
  if(from < 0) from = 0;
  if(nargs == 3)
  {
    if(!IsInteger(args[2]))
      LispError(5);   // Integer argument expected
    size = args[2]->ivalue;
    if(size < 0) size = -1;
  }
  int len = strlen(args[0]->svalue->text) - from;
  if((size > 0) && (len > size)) len = size;
  LISP result = NewString(len + 1);
  memcpy(result->svalue->text, args[0]->svalue->text + from, len);
  result->svalue->text[len] = '\0';
  return result;
}

// (SYMBOL <string>)
LISP Symbol(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]))
    LispError(8);   // String argument expected
  return(NewSym(args[0]->svalue->text));
}

// (STRING <value>) (STRING <list>)
LISP String(int nargs, LISP args[])
{
  LISP result;
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!args[0])
  {
    result = NewString(3);
    strcpy(result->svalue->text, "nil");
    return result;
  }
  if(IsSymbol(args[0]))
  {
    int len = args[0]->size - SYM_SIZE;
    result = NewString(len + 1);
    memcpy(result->svalue->text, args[0]->name, len);
    result->svalue->text[len] = '\0';
    return result;   
  }
  if(IsReal(args[0]) || IsInteger(args[0]))
  {
    char buf[40];
    if(IsReal(args[0]))
      sprintf(buf, ffmt, args[0]->rvalue);
    else
      sprintf(buf, ifmt, args[0]->ivalue);
    result = NewString(strlen(buf) + 1);
    strcpy(result->svalue->text, buf);
    return result;
  }
  if(IsCons(args[0]))
  {
    LISP c = args[0];
    char buf[40];
    int i;
    for(i = 0; (i < 40) && c; i++)
    {
      if(IsInteger(c->car) && (c->car->ivalue > 0) && (c->car->ivalue < 256))
      {
        buf[i] = (char)c->car->ivalue;
        c = c->cdr;
      }
      else
        LispError(25); // Illegal ASCII value
    }
    if(i == 40)
      LispError(26); // String too long
    buf[i] = '\0';
    result = NewString(strlen(buf) + 1);
    strcpy(result->svalue->text, buf);
    return result;
  }
  LispError(27); // Symbol, numeric or list argument expected
  return NIL;
}

// (CHARS <string>)
LISP Chars(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]))
    LispError(8);   // String argument expected
  PP* ppsave = PPLIST;
  volatile LISP s = args[0];
  PP ps;
  protect(&s, &ps);
  volatile LISP result = NIL;
  PP presult;
  protect(&result, &presult);
  for(int i = strlen(s->svalue->text) - 1; i >= 0; i--)
  {
    LISP c = NewInteger(s->svalue->text[i]);
    result = NewCons(c, result);
  }
  PPLIST = ppsave;
  return result;
}


// (STRING= <string> <string>)
LISP StringEqual(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]) || !IsString(args[1]))
    LispError(8);   // String argument expected
  if(strcmp(args[0]->svalue->text, args[1]->svalue->text))
    return NIL;
  else
    return T;
}

// (STRING> <string> <string>)
LISP StringGreater(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]) || !IsString(args[1]))
    LispError(8);   // String argument expected
  if(strcmp(args[0]->svalue->text, args[1]->svalue->text) > 0)
    return T;
  else
    return NIL;
}

// (STRING< <string> <string>)
LISP StringLess(int nargs, LISP args[])
{
  if(nargs != 2)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]) || !IsString(args[1]))
    LispError(8);   // String argument expected
  if(strcmp(args[0]->svalue->text, args[1]->svalue->text) < 0)
    return T;
  else
    return NIL;
}

// (TOKEN <string>)
LISP Token(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsString(args[0]))
    LispError(8);   // String argument expected
  char* sp = args[0]->svalue->text;
  char token[40];
  char* tp = token;
  LISP result = NIL;
  integer i = 0;
  while(isspace(*sp)) sp++, i++;
  if(!*sp) return result;
  if(isalpha(*sp))
  {
    // It's a symbol
    while(isalnum(*sp) && (i < 39)) { i++;  *tp++ = *sp++; };
    *tp = '\0';
    strupr(token);
    // Does this symbol exist?
    if(!strcmp(token, "NIL")) result = __EOF;
    else
    {
      result = FindSym(token);
      if(!result) result = NewSym(token);
    }
  }
  else if(isdigit(*sp))
  {
    // It's a number
    double d;
    int n;
    sscanf(sp, "%lf%n", &d, &n);
    i += n;
    result = NewReal(d);
  }
  else
  {
    // It's punctuation
    result = NewString(2);
    result->svalue->text[0] = *sp;
    result->svalue->text[1] = '\0';
    i += 1;
  }
  // Now lop off the bit we have scanned
  memmove(args[0]->svalue->text,
          args[0]->svalue->text + i,
          strlen(args[0]->svalue->text + i) + 1);
  return result;
}

// (LOAD '<file> [<verbose>])
LISP Load(int nargs, LISP args[])
{
  if(nargs < 1 || nargs > 2)
    LispError(2);   // Incorrect number of args
  // Open the filename given in a symbol or string
  FILE* f = 0;
  if(IsSymbol(args[0]))
  {
    strncpy(Filename, args[0]->name, args[0]->size - SYM_SIZE);
    Filename[args[0]->size - SYM_SIZE] = '\0';
    f = fopen(Filename, "r");
  }
  else if(IsString(args[0]))
    f = fopen(args[0]->svalue->text, "r");
  else
    LispError(8);   // String argument expected
  if(!f) return NIL;        // Return NIL if open fails
  // Check verbose flag
  bool verbose = ((nargs == 2) && (args[1]->value));
  // Read and evaluate expressions making sure that we close the file
  // whatever happens
  try
  {
    while(!feof(f))
    {
      LISP result = Eval(read(f));
      if(verbose)
      {
        print(result, stdout, false);
        printf("\n");
      }
    }
  }
  catch(...)       // Make sure the file is closed
  {
    fclose(f);
    throw;
  }
  fclose(f);       // Normal exit
  return T;
}

// (COLLECT)
LISP Collect(int nargs, LISP [])
{
  if(nargs)
    LispError(2);   // Incorrect number of args
  GC();
  return NIL;
}

// (ERROR <string> [<num>])
LISP _Error(int nargs, LISP args[])
{
  if((nargs < 1) || (nargs > 2))
    LispError(2);   // Incorrect number of args
  if((nargs == 1) && args[0] && IsInteger(args[0]))
    LispError(args[0]->ivalue);
  if(args[0] && !IsString(args[0]))
    LispError(8);   // String argument expected
  if(nargs == 2)
  {
    if(IsInteger(args[1]))
      LispError(args[1]->ivalue, (args[0]) ? args[0]->svalue->text : NULL);
    else LispError(5);   // Integer argument expected
  }
  LispError(1, args[0]->svalue->text);
  return NIL;
}

// (ERRORSET <body> ...)
LISP Errorset(LISP args)
{
  PP* ppsave = PPLIST;
  PFrame savebase = base;
  LISP result;
  try
  {
    try
    {
      result = List(args);
    }
    catch(Error e)
    {
      // Don't trap fatal errors
      if(e.err < 0) throw;
      // Restore the protected pointer list
      PPLIST = ppsave;
      // Restore our stack frame
      while(base != savebase) PopFrame();
      result = (NewInteger(e.err));
    }
  }
  __except(FilterException(GetExceptionCode()))
  {
    Error aritherror = OSException(GetExceptionCode());
    if(aritherror.err)
    {
      PPLIST = ppsave;
      // Restore our stack frame
      while(base != savebase) PopFrame();
      result = (NewInteger(aritherror.err));
    }
  }
  return result;
}

// (BACKTRACE <integer>)
LISP Backtrace(int nargs, LISP args[])
{
  if(nargs != 1)
    LispError(2);   // Incorrect number of args
  if(!IsInteger(args[0]))
    LispError(5);   // Integer argument expected
  backtrace = args[0]->ivalue;
  return args[0];
}

//---------------------------------------------------------------------------


