//---------------------------------------------------------------------------
#include "AJBLisp.h"

//---------------------------------------------------------------------------
char ifmt[10] = "%1ld";
char ffmt[10] = "%#1.8lg";

const int ibufsz = 241;

LISP ratom(FILE* in)
{
  char buffer[ibufsz];             // Input buffer
  char* p = buffer;                // Buffer pointer
  int c;

  c = fgetc(in);
  if(c == '#')                     // Read macro
  {
    c = fgetc(in);
    if(c == '\\')                  // c reads character c as its ASCII value
    {
      c = fgetc(in);
      if(c != EOF)
        return(NewInteger(c));
    }
    LispError(48); // Illegal read macro
  }

  while(true)
  {
    // End of symbol?
    if(isspace(c) || (c == '(') || (c == ')') || (c == EOF) || (c == ';'))
      break;
    if(c == '\\')     // Escape character
    {
      c = fgetc(in);
      if(c == EOF)
        LispError(40); // EOF after reading escape character
    }
    *p++ = (char)c;
    if((p - buffer) == ibufsz)
      LispError(47);
    c = fgetc(in);
  }
  if(c != EOF) ungetc(c, in);
  *p = '\0';
  strupr(buffer);
  // Now the symbol is in the buffer determine if it is an integer
  int i, j;
  integer n;    // The number of characters converted
  j = sscanf(buffer, "%d%n", &i, &n);
  if((j == 1) && (n == strlen(buffer)))
    return(NewInteger(i));
  // Maybe it is a real
  double r;
  j = sscanf(buffer, "%lf%n", &r, &n);
  if((j == 1) && (n == strlen(buffer)))
    return(NewReal(r));
  // Must be a symbol
  if(!strcmp(buffer, "NIL")) return NIL;
  // Does this symbol exist?
  LISP s = FindSym(buffer);
  if(s) return s;
  // No, create a new one
  return NewSym(buffer);
}

// Read a string
LISP rstring(FILE* in)
{
  PP* ppsave = PPLIST;
  volatile LISP s = NewString(0);
  PP ps;
  protect(&s, &ps);
  int size = 0;
  int c;
  do
  {
    c = fgetc(in);
    if(c == EOF)
      LispError(41); // EOF inside string
    if(c < ' ')
      LispError(42); // Illegal character inside string
    if(c == '\"') c = '\0';
    if(c == '\\')
    {
      c = fgetc(in);
      if(c == EOF)
        LispError(41); // EOF inside string
      if(c == 'n') c = '\n';
      if(c == 'r') c = '\r';
    }
    // Now store the character into the string
    if(shfree >= shlimit)
    {
      GC();
      if(shfree >= shlimit)
        LispError(17); // Out of string space
    }
    *shfree++ = (char)c;
    s->svalue->size = SCELL_SIZE + (++size);
  } while(c);
  PPLIST = ppsave;
  return s;
}

int skipspace(FILE* in)
{
  int c;
  do
  {
    c = fgetc(in);
    if(c == EOF) return EOF;
    if(c == ';')
    {
      do
      {
        c = fgetc(in);
        if(c == EOF) return EOF;
      } while((c != '\n') && (c != '\r'));
    }
  } while(isspace(c));
  return c;
}

// Read an s-expression
LISP read(FILE* in)
{
  LISP temp;
  int c;
  c = skipspace(in);
  switch(c)
  {
    case EOF:
      return __EOF;
    case '(':
      return(readlist(in));
    case ')':
      LispError(43); // Unexpected close parenthesis
    case '\'':
      temp = read(in);
      temp = NewCons(temp, NIL);
      return NewCons(QUOTE, temp);
    case '"':
      return rstring(in);
    default:
      ungetc(c, in);
      return ratom(in);
  }
}

LISP readlist(FILE* in)
{
  int c;
  c = skipspace(in);
  if(c == EOF)
    LispError(44); // EOF inside list
  if (c == ')')
    return(NIL);
  ungetc(c, in);
  PP* ppsave = PPLIST;
  volatile LISP t = NIL;
  PP pt;
  protect(&t, &pt);
  t = read(in);
  if(t == DOT)
  {
    t = read(in);
    c = skipspace(in);
    if(c == EOF)
      LispError(44); // EOF inside list
    if (c != ')')
      LispError(45); // Missing close parenthesis
    PPLIST = ppsave;
    return(t);
  }
  LISP temp = readlist(in);
  PPLIST = ppsave;
  return NewCons(t,temp);
}

LISP flush(FILE* in)
{
  fflush(in);
  return NIL;
}

// Read input line verbatim as a string
LISP readline(FILE* in)
{
  if(in == stdin) fflush(stdin);
  PP* ppsave = PPLIST;
  volatile LISP s = NewString(0);
  PP ps;
  protect(&s, &ps);
  int size = 0;
  int c;
  do
  {
    c = fgetc(in);
    if((c == EOF) || (c == '\n') || (c == '\r'))
      c = '\0';
    // Now store the character into the string
    if(shfree >= shlimit)
    {
      GC();
      if(shfree >= shlimit)
        LispError(17); // Out of string space
    }
    *shfree++ = (char)c;
    s->svalue->size = SCELL_SIZE + (++size);
  } while(c);
  PPLIST = ppsave;
  return s;
}

// patom() outputs a printable representation of an atom to a stream
// The output formats are as follows
// Symbol      Output name
// String      Output string value
// Integer     Output value in text form
// Real        Output value in text form
// Subr        Output as Subr#<address>
// FSubr       Output as FSubr#<address>
// File        Output as File#<address>
// <address> is the execution address of the code as eight hex digits
// e.g. Subr#002A37F5
// This is really only useful for debugging the interpreter.
// (Why else are you printing F/Subr atoms anyway?)
// Subrs, FSubrs and Files cannot be read back in.
LISP patom(LISP c, FILE* out, bool cflag)
{
  if(!c)
    fputs("NIL", out);
  else
    switch(c->flags)
    {
      case tCONS:
        LispError(46); // List argument to patom
    //break;
      case tSYM:
        if(cflag)
          fwrite(c->name, 1, c->size - SYM_SIZE, out);
        else
        {
          for(integer i = 0; i < (c->size - SYM_SIZE); i++)
          {
            char ch = c->name[i];
            if(isspace(ch) || (ch == '(') || (ch == ')') || (ch == ';') ||
               ((i == 0) && (ch == '\''))) fputc('\\', out);
            fputc(ch, out);
          }
        }
      break;
      case tSTR:
        if(!cflag) fputc('"', out);
        fputs(c->svalue->text, out);
        if(!cflag) fputc('"', out);
      break;
      case tINT:
        fprintf(out, ifmt, c->ivalue);
      break;
      case tREAL:
        fprintf(out, ffmt, c->rvalue);
      break;
      case tSUBR:
        fprintf(out, "Subr#%p", c->code);
      break;
      case tFSUBR:
        fprintf(out, "FSubr#%p", c->code);
      break;
      case tFILE:
        fprintf(out, "File#%p", c->code);
      break;
    }
  return c;
}

LISP print(LISP c, FILE* out, bool cflag)
{
  if(!c || IsAtom(c))
    return patom(c, out, cflag);
  fputc('(', out);
  printlist(c, out, cflag);
  fputc(')', out);
  return(c);
}

void printlist(LISP c, FILE* out, bool cflag)
{
  while(true)
  {
    print(c->car, out, cflag);
    c = c->cdr;
    if(IsAtom(c))
    {
      if(c)
      {
        fputs(" . ", out);
        patom(c, out, cflag);
      }
      return;
    }
    fputc(' ', out);
  }
}

//---------------------------------------------------------------------------



