//---------------------------------------------------------------------------
#include "AJBLisp.h"

//---------------------------------------------------------------------------

// Note: All heap references are even addresses.  This leaves the door open
//       to use odd addresses to represent 31-bit integers.

// Globals relating to the heap
pointer  heap;
integer  hsize;
pointer  hfree;
pointer  hlimit;
pointer  gcheap;
pointer  gcfree;
LISP     hashtable[HASHSIZE];

// Globals relating to the stack
pointer  stack;     // Base of allocated stack memory
pointer  slimit;    // End of allocated stack memory
pointer  stop;      // Next free entry
PFrame   base;      // Base of current frame

// Globals relating to the string heap
pointer  sheap;     // Base of allocated memory
pointer  shlimit;   // End of allocated memory
pointer  shfree;    // Next free entry

// Protected pointer list
PP* PPLIST = 0;

// Special lisp atoms
const LISP NIL = (LISP) 0;     // Nil, nada, nothing
LISP UNDEFINED;                // Initial value of symbols
LISP T;                        // Truth
LISP LAMBDA;                   // EXPR marker
LISP DOT;                      // The CONS operator
LISP __EOF;                    // End of file
LISP QUOTE;                    // The QUOTE FSUBR
// Backtrace frame limit
integer backtrace = 0;
//---------------------------------------------------------------------------

void Heap(integer size)
{
  gcheap = NULL;
  heap = new byte[size];
  if(!heap)
    LispError(-1); // Failed to allocate heap
  hsize = size;
  hlimit = heap + size;
  hfree = heap;
  HeapInit();
}

void HeapInit()
{
  // Initialise special atoms
  // The values must be set explicitly because Set()
  // requires QUOTE to be defined and it will not change
  // special symbols in any case
  UNDEFINED = NewSym("*UNDEFINED*");
  UNDEFINED->value = UNDEFINED;

  T = NewSym("T");
  T->value = T;

  LAMBDA = NewSym("LAMBDA");
  LAMBDA->value = LAMBDA;

  DOT = NewSym(".");
  DOT->value = DOT;

  __EOF = NewSym("*EOF*");
  __EOF->value = __EOF;

  QUOTE = NewSym("QUOTE");
  QUOTE->value = NewFSubr(Quote);

  // The remainder can be installed more easily
  _Set(NewSym("SPACE"),          NewSym(" "));     // For use with PRINT
  _Set(NewSym("CR"),             NewSym("\n"));    // For use with PRINT
  _Set(NewSym("EQ"),             NewSubr(Eq));
  _Set(NewSym("EQL"),            NewSubr(Eql));
  _Set(NewSym("EQUAL"),          NewSubr(Equal));
  _Set(NewSym("CONS"),           NewSubr(Cons));
  _Set(NewSym("NOT"),            NewSubr(Not));
  _Set(NewSym("NULL"),           NewSubr(Not));
  _Set(NewSym("AND"),            NewFSubr(And));
  _Set(NewSym("OR"),             NewFSubr(Or));
  _Set(NewSym("BNOT"),           NewSubr(Bnot));
  _Set(NewSym("BAND"),           NewSubr(Band));
  _Set(NewSym("BOR"),            NewSubr(Bor));
  _Set(NewSym("SETQ"),           NewFSubr(Setq));
  _Set(NewSym("SET"),            NewSubr(Set));
  _Set(NewSym("CAR"),            NewSubr(Car));
  _Set(NewSym("CDR"),            NewSubr(Cdr));
  _Set(NewSym("CAAR"),           NewSubr(Caar));
  _Set(NewSym("CADR"),           NewSubr(Cadr));
  _Set(NewSym("CDAR"),           NewSubr(Cdar));
  _Set(NewSym("CDDR"),           NewSubr(Cddr));
  _Set(NewSym("RPLACA"),         NewSubr(Rplaca));
  _Set(NewSym("RPLACD"),         NewSubr(Rplacd));
  _Set(NewSym("EVAL"),           NewSubr(_Eval));
  _Set(NewSym("LENGTH"),         NewSubr(Length));
  _Set(NewSym("+"),              NewSubr(Add));
  _Set(NewSym("-"),              NewSubr(Subtract));
  _Set(NewSym("*"),              NewSubr(Multiply));
  _Set(NewSym("/"),              NewSubr(Divide));
  _Set(NewSym("REM"),            NewSubr(Rem));
  _Set(NewSym(">"),              NewSubr(Greaterp));
  _Set(NewSym("<"),              NewSubr(Lessp));
  _Set(NewSym("<="),             NewSubr(NGreaterp));
  _Set(NewSym(">="),             NewSubr(NLessp));
  _Set(NewSym("="),              NewSubr(Equalp));
  _Set(NewSym("<>"),             NewSubr(NEqualp));
  _Set(NewSym("ZEROP"),          NewSubr(Zerop));
  _Set(NewSym("ONEP"),           NewSubr(Onep));
  _Set(NewSym("MINUSP"),         NewSubr(Minusp));
  _Set(NewSym("ATOM"),           NewSubr(Atom));
  _Set(NewSym("SYMBOLP"),        NewSubr(Symbolp));
  _Set(NewSym("NUMBERP"),        NewSubr(Numberp));
  _Set(NewSym("LISTP"),          NewSubr(Listp));
  _Set(NewSym("STRINGP"),        NewSubr(Stringp));
  _Set(NewSym("SUBRP"),          NewSubr(Subrp));
  _Set(NewSym("FSUBRP"),         NewSubr(FSubrp));
  _Set(NewSym("IF"),             NewFSubr(If));
  _Set(NewSym("COND"),           NewFSubr(Cond));
  _Set(NewSym("LOOP"),           NewFSubr(Loop));
  _Set(NewSym("UNTIL"),          NewFSubr(Until));
  _Set(NewSym("WHILE"),          NewFSubr(While));
  _Set(NewSym("PROGN"),          NewFSubr(Progn));
  _Set(NewSym("LET"),            NewFSubr(Let));
  _Set(NewSym("SLET"),           NewFSubr(Slet));
  _Set(NewSym("LIST"),           NewFSubr(List));
  _Set(NewSym("MEMBER"),         NewSubr(Member));
  _Set(NewSym("OPEN"),           NewSubr(Open));
  _Set(NewSym("CLOSE"),          NewSubr(Close));
  _Set(NewSym("FLOAT-FORMAT"),   NewSubr(FloatFormat));
  _Set(NewSym("INTEGER-FORMAT"), NewSubr(IntegerFormat));
  _Set(NewSym("PRIN"),           NewFSubr(Prin));
  _Set(NewSym("PRINC"),          NewFSubr(Princ));
  _Set(NewSym("PRINT"),          NewFSubr(Print));
  _Set(NewSym("PRINTC"),         NewFSubr(Printc));
  _Set(NewSym("READ"),           NewSubr(Read));
  _Set(NewSym("READLINE"),       NewSubr(Readline));
  _Set(NewSym("FLUSH"),          NewSubr(Flush));
  _Set(NewSym("ABS"),            NewSubr(Abs));
  _Set(NewSym("FLOAT"),          NewSubr(Float));
  _Set(NewSym("INTEGER"),        NewSubr(Integer));
  _Set(NewSym("FLOOR"),          NewSubr(Floor));
  _Set(NewSym("CEIL"),           NewSubr(Ceil));
  _Set(NewSym("SQRT"),           NewSubr(Sqrt));
  _Set(NewSym("SIN"),            NewSubr(Sin));
  _Set(NewSym("COS"),            NewSubr(Cos));
  _Set(NewSym("TAN"),            NewSubr(Tan));
  _Set(NewSym("ASIN"),           NewSubr(Asin));
  _Set(NewSym("ACOS"),           NewSubr(Acos));
  _Set(NewSym("ATAN"),           NewSubr(Atan));
  _Set(NewSym("LOG"),            NewSubr(Log));
  _Set(NewSym("LN"),             NewSubr(Ln));
  _Set(NewSym("EXP"),            NewSubr(Exp));
  _Set(NewSym("POWER"),          NewSubr(Power));
  _Set(NewSym("RANDOM"),         NewSubr(Random));
  _Set(NewSym("ASSOC"),          NewSubr(Assoc));
  _Set(NewSym("PUT"),            NewSubr(Put));
  _Set(NewSym("GET"),            NewSubr(Get));
  _Set(NewSym("REMPROP"),        NewSubr(Remprop));
  _Set(NewSym("PLIST"),          NewSubr(Plist));
  _Set(NewSym("OBLIST"),         NewSubr(Oblist));
  _Set(NewSym("STRLEN"),         NewSubr(Strlen));
  _Set(NewSym("CONCAT"),         NewSubr(Concat));
  _Set(NewSym("SUBSTRING"),      NewSubr(Substring));
  _Set(NewSym("SYMBOL"),         NewSubr(Symbol));
  _Set(NewSym("STRING"),         NewSubr(String));
  _Set(NewSym("CHARS"),          NewSubr(Chars));
  _Set(NewSym("STRING="),        NewSubr(StringEqual));
  _Set(NewSym("STRING>"),        NewSubr(StringGreater));
  _Set(NewSym("STRING<"),        NewSubr(StringLess));
  _Set(NewSym("TOKEN"),          NewSubr(Token));
  _Set(NewSym("LOAD"),           NewSubr(Load));
  _Set(NewSym("COLLECT"),        NewSubr(Collect));
  _Set(NewSym("ERROR"),          NewSubr(_Error));
  _Set(NewSym("ERRORSET"),       NewFSubr(Errorset));
  _Set(NewSym("BACKTRACE"),      NewSubr(Backtrace));
}

void FreeHeap()
{
  if(heap)   delete [] heap;
  if(gcheap) delete [] gcheap;
}

LISP HeapAlloc(integer size)
{
  pointer c = hfree;
  hfree += size;
  if(hfree > hlimit)
  {
    GC();
    c = hfree;
    hfree += size;
    if(hfree > hlimit)
      LispError(-3); // Out of heap space
  }
  return (LISP)c;
}

LISP NewCons(LISP car, LISP cdr)
{
  PP* ppsave = PPLIST;
  volatile LISP pcar = car;
  PP ppcar;
  protect(&pcar, &ppcar);
  volatile LISP pcdr = cdr;
  PP ppcdr;
  protect(&pcdr, &ppcdr);
  LISP c = HeapAlloc(CONS_SIZE);
  c->flags = tCONS;
  c->size = CONS_SIZE;
  c->cdr = pcdr;
  c->car = pcar;
  PPLIST = ppsave;
  return c;
}

LISP NewSym(char* str)
{
  int len = strlen(str);
  if(len > 240)
    LispError(18);       // Symbol name too long
  int alloc = len + SYM_SIZE;
  if(alloc & 1) alloc++; // Ensure word alignment
  LISP c = HeapAlloc(alloc);
  c->flags = tSYM;
  c->size = (byte)(len + SYM_SIZE);
  c->value = UNDEFINED;
  c->prop = NULL;
  // Link symbol into hashchain
  int hashidx = Hash(str);
  c->hash = hashtable[hashidx];
  hashtable[hashidx] = c;
  memcpy(c->name, str, len);
  return c;
}

LISP NewString(int len)
{
  LISP c = HeapAlloc(STR_SIZE);
  c->flags = tSTR;
  c->size = STR_SIZE;
  c->svalue = AllocString(c, len);
  return c;
}

LISP NewInteger(int val)
{
  LISP c = HeapAlloc(INT_SIZE);
  c->flags = tINT;
  c->size = INT_SIZE;
  c->ivalue = val;
  return c;
}

LISP NewReal(double val)
{
  LISP c = HeapAlloc(REAL_SIZE);
  c->flags = tREAL;
  c->size = REAL_SIZE;
  c->rvalue = val;
  return c;
}

LISP NewSubr(SUBR code)
{
  LISP c = HeapAlloc(EXEC_SIZE);
  c->flags = tSUBR;
  c->size = EXEC_SIZE;
  c->code = (pointer)code;
  return c;
}

LISP NewFSubr(FSUBR code)
{
  LISP c = HeapAlloc(EXEC_SIZE);
  c->flags = tFSUBR;
  c->size = EXEC_SIZE;
  c->code = (pointer)code;
  return c;
}

// Return symbol given name or NULL if not found
LISP FindSym(char* name)
{
  LISP c = hashtable[Hash(name)];
  byte len = (byte)strlen(name);
  byte clen = (byte)(SYM_SIZE + len);
  while(c)
  {
    // Length test quickly rejects most mismatches
    if((c->size == clen) && (!strncmp(name, c->name, len)))
      break;
    // Chase the hash-chain
    c = c->hash;
  }
  return c;
}

void GC()
{
  // printf("gc\n");
  // Grab this before it moves
  LISP undefined = UNDEFINED;
  // Create a new heap
  gcheap = new byte[hsize];
  gcfree = gcheap;
  // Move these statically referenced cells first
  MoveCell(&UNDEFINED);
  MoveCell(&T);
  MoveCell(&LAMBDA);
  MoveCell(&DOT);
  MoveCell(&__EOF);
  MoveCell(&QUOTE);
  // Move all protected cells
  PP* p = PPLIST;
  while(p)
  {
    MoveCell((LISP*)p->pp);
    p = p->next;
  }
  // Call MoveCell for every CSYM that is defined or has property
  for(int i = 0; i < HASHSIZE; i++)
  {
    LISP next = hashtable[i];
    while(next)
    {
      if((next->value != undefined) || (next->prop))
        MoveCell(&next);
      next = next->hash;
    }
  }
  // Move stacked references
  SweepStack();
  // Move referenced cells
  Sweep();
  // Fixup the hash chains
  FixHash();
  // Collect dead strings
  SGC();
  // Prove that all cells have been moved
  //  memset(heap, 0, hsize);
  // We are done with the old heap
  delete [] heap;
  // Install the new heap
  heap = gcheap;
  hfree = gcfree;
  hlimit = gcheap + hsize;
  gcheap = NULL;
}

// If cell is in the Heap then move it to the GCHeap
void MoveCell(LISP* cell)
{
  if(!*cell) return;
  if(IsMoved(*cell))
    // Update reference
    *cell = (*cell)->reloc;
  else
  {
    // Copy cell to GCHeap
    memcpy(gcfree, *cell, (*cell)->size);
    // Leave a forwarding address
    (*cell)->reloc = (LISP)gcfree;
    (*cell)->flags |= MOVED;
    (*cell) = (LISP)gcfree;
    // Update gcfree
    gcfree += (*cell)->size;
    // Keep word alignment (CSYMs can upset this)
    if((integer)gcfree & 1) gcfree++;
  }
}

// Sweep the new heap for unmoved references
// A single sweep is enough since referenced cells are moved to higher
// addresses and will be swept after they have been moved
void Sweep()
{
  pointer c = gcheap;
  while((pointer)c < gcfree)
  {
    // Move referenced cells
    if(IsCons((LISP)c))
    {
      MoveCell(&((LISP)c)->car);
      MoveCell(&((LISP)c)->cdr);
    }
    else if(IsSymbol((LISP)c))
    {
      MoveCell(&((LISP)c)->value);
      MoveCell(&((LISP)c)->prop);
      // Hash chains are fixed up later
    }
    // Next cell
    c += ((LISP)c)->size;
    // Keep word alignment
    if((integer)c & 1) c++;
  }
}

// Rebuild the hash chains excluding dead cells
void FixHash()
{
  for(int i = 0; i < HASHSIZE; i++)
  {
    LISP next = hashtable[i];
    // Remove dead cells at head of hash chain
    while(next && !IsMoved(next))
      next = next->hash;
    // Is this chain now empty?
    if(!next)
    {
      // Yup!
      hashtable[i] = NULL;
      continue;
    }
    // hashend is the last cell of the new hashchain
    LISP hashend = hashtable[i] = next->reloc;
    // Scan next through the old hashchain
    while(next)
    {
      // If a valid cell is found
      LISP link = next->hash;
      if(next && IsMoved(next))
      {
        // Link the relocated cell to the new hashchain
        hashend = hashend->hash = next->reloc;
      }
      next = link;
    }
    // Terminate the new hashchain
    hashend->hash = NULL;
  }
}

int Hash(char* s)
{
  // Distribution has been tested - looks fine
  integer hash = strlen(s);
  while(*s)
    hash = (5 * hash) ^ ((integer)(*s++));
  return hash % HASHSIZE;
}

// Not used - This is a debugging function which dumps the heap to stdout
// If I delete it then my next edit will almost certainly cause heap corruption
// (Murphey's Law), so it's still here
void DumpHeap()
{
  pointer c = heap;
  while(c < hfree)
  {
    print(((LISP)c), stdout, false);
    if(IsSymbol((LISP)c))
    {
      printf(" Value: ");
      print(((LISP)c)->value, stdout, false);
      printf(" Property: ");
      print(((LISP)c)->prop, stdout, false);
    }
    printf("\n");
    void getchar();
    // Next cell
    c += ((LISP)c)->size;
    if((int)c & 1) c++;
  }
}


//---------------------------------------------------------------------------
void Stack(integer size)
{
  stack = new byte[size];
  if(!stack)
    LispError(-2); // Failed to allocate stack
  slimit = stack + size;
  stop = stack;
  base = (PFrame)stack;
}

void FreeStack()
{
  delete [] stack;
}

void NewFrame(LISP expr)
{
  if((slimit - stop) < 8)
    LispError(-4);                            // Out of stack space
  PFrame frame = (PFrame)stop;
  frame->last = (word)(stop - (pointer)base); // Set offset to prior frame
  base = frame;                               // New frame base
  frame->nargs = 0;                           // No args (yet)
  frame->save = false;                        // Arg frame initially
  frame->fun = expr;                          // Pointer to sexpr
  stop += sizeof(Frame);                      // Update stack top
}

// Pop a frame and restore the environment
void PopFrame()
{
  // Arg frames can be popped without further consideration
  // but Save frames must be restored
  if(base->save)
  {
    // We must now restore the previous environment
    int argp = 0;                          // Index into array of saved args
    for(int n = 0; n < base->nargs; n++)
    {
      LISP arg = base->args[argp++];       // Arg is the cell on the heap
      // This and the remaing args are evaluated but unbound
      if(!arg) break;                      // Unbound argument
      if(!IsSymbol(arg))
        LispError(-5);                     // Corrupted lisp stack
      arg->value = base->args[argp++];     // Restore the old value
    }
  }
  stop = (pointer)base;                    // Restore old stack top
  base = (PFrame)(stop - base->last);      // Restore old frame base
}

// Push an arg onto the current frame
// This will protect the stacked pointer from GC
void Push(LISP c)
{
  if(stop == slimit)
    LispError(-4);            // Out of stack space
    *(LISP*)stop = c;
  stop += sizeof(LISP);
}

// Pop arg from the current frame
// Returns the arg popped
LISP Pop()
{
  stop -= sizeof(LISP);
  return *(LISP*)stop;
}

// Evaluate a list of args onto the stack
void EvList(LISP fargs)
{
  PP* ppsave = PPLIST;
  volatile LISP pfargs = fargs;
  PP ppfargs;
  protect(&pfargs, &ppfargs);

  // If this is an arg frame just evaluate each arg and push it
  // If this is a save frame create saved arg entries
  // which are NULL but for the saved value which is the
  // evaluated arg
  // Bind() will then bind these values to the arguments
  // and store the old environment in the space allocated here
  bool save = base->save;
  while(pfargs)
  {
    LISP evarg = Eval(pfargs->car);
    if(save) Push(NIL);       // Space for symbol pointer
    Push(evarg);
    base->nargs++;
    if(base->nargs == 0)      // Oops! we wrapped around
      LispError(19);          // Too many arguments for stack frame
    pfargs = pfargs->cdr;
  }
  PPLIST = ppsave;
}

// Bind a list of formal args to their evaluated values
// in the frame and save the old environment into the frame
void Bind(LISP formals)
{
  LISP* args = base->args;
  LISP formal;
  int nargs = base->nargs;
  while(formals && nargs)
  {
    formal = formals->car;
    formals = formals->cdr;
    // Handle list (defaulted) formal arguments
    if(IsCons(formal)) formal = formal->car;
    if(!IsSymbol(formal))
      throw("Illegal formal argument in EXPR");
    *args++ = formal;
    // Get current value
    LISP oldval = formal->value;
    // Store new value
    formal->value = *args;
    // Save the old value
    *args++ = oldval;
    nargs --;
  }
  if(nargs)
    LispError(2);             // Incorrect number of arguments to function
  if(!formals) return;
  // Now deal with any optional arguments
  PP* ppsave = PPLIST;
  volatile LISP pformals = formals;
  PP ppformals;
  protect(&pformals, &ppformals);
  while(pformals)
  {
    formal = pformals->car;
    if(!IsCons(formal))
      LispError(2);           // Incorrect number of arguments to function
    if(!IsSymbol(formal->car))
      LispError(12);          // Illegal formal argument
    Push(formal->car);
    Push(formal->car->value);
    LISP val = Eval(formal->cdr);
    // Note: use pformals->car because formal may be moved by Eval
    pformals->car->car->value = val;
    base->nargs++;
    if(base->nargs == 0)      // Oops! we wrapped around
      LispError(19);          // Too many arguments for stack frame
    pformals = pformals->cdr;
  }
  PPLIST = ppsave;
}

// Sweep the stack moving all cells referenced to new heap
// Also relocates protected pointers
//
// The Stack frames look like this:
//
//   +-------------------+
//   | saved arg         | ----> CELL
//   | saved arg         | ----> CELL
//   | ...               |
//   | function pointer  | ----> CELL
//   +-------------------+
//   | frame header      |
//   +-------------------+
//   Every line of the above diagram is a longword
//   The frame header should not be adjusted at all
//   The saved args should be moved
//   The size of the saved args block is:
//   (nargs + 1) * sizeof(LISP) for an arg frame and
//   (nargs * 2 + 1) * sizeof(LISP) for a save frame
//   The +1 includes the function pointer
void SweepStack()
{
  if(stop == stack) return;   // Nothing to do
  PFrame sbase = base;
  while(true)
  {
    int n = sbase->nargs;
    if(sbase->save) n <<= 1;
    // Move cells on stack
    // Note: sbase->args[-1] is sbase->fun
    for(int i = -1; i < n; i++)
      MoveCell(&sbase->args[i]);
    // Last is zero in the root frame
    if(sbase->last == 0) break;
    // Step back to previous frame
    sbase = (PFrame)((pointer)sbase - sbase->last);
  }
}

void ClearStack()
{
  PPLIST = 0;                 // Don't attempt to access dead stack!
  if(backtrace > 0) printf("Backtrace\n");
  int bt = backtrace;
  while(stop > stack)
  {
    if(bt-- > 0)
    {
      printf("Eval:  ");
      print(base->fun, stdout, false);
      printf("\n");
      for(int i = 0; i < base->nargs; i++)
      {
        if(base->save)
        {
          printf("  Arg%-2d: ", i+1);
          if(!base->args[i << 1])
            printf("UNBOUND\n");
          else
          {
            print(base->args[i << 1], stdout, false);
            printf(" = ");
            print(base->args[i << 1]->value, stdout, false);
            printf("\n");
          }
        }
        else
        {
          printf("  Arg:  ");
          print(base->args[i], stdout, false);
          printf("\n");
        }
      }
    }
    PopFrame();
  }
}
//---------------------------------------------------------------------------
void SHeap(integer size)
{
  sheap = new byte[size];
  if(!sheap)
    LispError(-6);            // Failed to allocate string heap
  shlimit = sheap + size;
  shfree = sheap;
}
//---------------------------------------------------------------------------
STRING AllocString(LISP c, int len)
{
  PP* ppsave = PPLIST;
  volatile LISP pc = c;
  PP ppc;
  protect(&pc, &ppc);
  len += SCELL_SIZE;
  if((shfree + len) >= shlimit)
  {
    GC();
    if((shfree + len) >= shlimit)
      LispError(17);          // Out of string space
  }
  STRING s = (STRING)shfree;
  shfree += len;
  s->cell = pc;
  s->size = len;
  s->text[0] = '\0';
  PPLIST = ppsave;
  return s;
}
//---------------------------------------------------------------------------
// Garbage collect string heap
void SGC()
{
  pointer sptr = sheap;       // pointer to next STRCELL
  pointer newfree = sheap;    // pointer to new free space
  while(sptr < shfree)
  {
    int size = ((STRING)sptr)->size;
    // Is this STRCELL is still referenced?
    if(IsMoved(((STRING)sptr)->cell))
    {
      if(sptr != newfree)     // Is there free space to move into?
        memmove(newfree, sptr, size);
      // Update cell pointer in STRCELL
      ((STRING)newfree)->cell = ((STRING)newfree)->cell->reloc;
      // Update STRING pointer in CELL
      ((STRING)newfree)->cell->svalue = (STRING)newfree;
      newfree += size;
    }
    sptr += size;
  }
  shfree = newfree;
}
//---------------------------------------------------------------------------
void LispError(int err, char *msg)
{
  Error e;
  e.err = err;
  e.msg = msg;
  throw e;
}
//---------------------------------------------------------------------------



