//---------------------------------------------------------------------------
#ifndef StoreH
#define StoreH

typedef unsigned char   byte;
typedef unsigned short  word;
typedef unsigned long   integer;
typedef byte*           pointer;

#define HASHSIZE 79

// Backtrace frame limit
extern integer backtrace;

enum TYPE {tCONS, tSYM, tSTR, tINT, tREAL, tSUBR, tFSUBR, tFILE};
const byte MOVED = 0x80;

const integer CONS_SIZE = 10;
const integer SYM_SIZE  = 14; // + size of name
const integer STR_SIZE  = 6;
const integer REAL_SIZE = 10;
const integer INT_SIZE  = 6;
const integer EXEC_SIZE = 6;
const integer FILE_SIZE = 6;
const integer SCELL_SIZE = 8;

struct CELL;
typedef CELL*  LISP;

// The protected pointer chain
struct PP
{
  PP*  next;
  volatile LISP* pp;
};

extern PP* PPLIST;
inline void protect(volatile LISP* l, PP* p)
{
  p->pp = l;
  p->next = PPLIST;
  PPLIST = p;
}

// These are used to typecast F/Subr code pointers
// It is a pointer to a function returning a lisp cell
// Subrs take an arg count and an arg array
// Fsubrs take the unevaluated arglist
typedef LISP (*SUBR)(int, LISP[]);
typedef LISP (*FSUBR)(LISP);

struct STRCELL
{
  LISP    cell;               // The string cell on the lisp heap
  integer size;               // The size of this STRCELL
  char    text[];             // The null terminated string
};

typedef STRCELL* STRING;

#pragma anon_struct on

struct CELL
{
  byte         flags;
  byte         size;
  union
  {
    struct
    {
      LISP     car;
      LISP     cdr;
    };
    struct
    {
      LISP     value;
      LISP     prop;
      LISP     hash;
      char     name[];
    };
    STRING     svalue;
    double     rvalue;
    int        ivalue;
    pointer    code;
    FILE*      file;
    LISP       reloc;
  };
};

#pragma anon_struct off

// Globals relating to the heap
extern pointer  heap;
extern integer  hsize;
extern pointer  hfree;
extern pointer  hlimit;
extern pointer  gcheap;
extern pointer  gcfree;
extern LISP     hashtable[HASHSIZE];

// Prototypes for heap related functions
void Heap(integer size);
void HeapInit();
void FreeHeap();
LISP HeapAlloc(integer size);
LISP NewCons(LISP car, LISP cdr);
LISP NewSym(char* str);
LISP NewString(int len);
LISP NewInteger(int val);
LISP NewReal(double val);
LISP NewSubr(SUBR code);
LISP NewFSubr(FSUBR code);
LISP FindSym(char* name);
LISP Set(LISP c, LISP value);
void GC();
void MoveCell(LISP* cell);
void Sweep();
void FixHash();
int Hash(char* s);
void DumpHeap();

// Utility functions
inline bool IsMoved(LISP c)  // NIL is not a valid argument
{
  return (bool)((c->flags & MOVED) != 0);
}

inline bool IsCons(LISP c)
{
  return (bool)(c && (c->flags == tCONS));
}

inline bool IsAtom(LISP c)
{
  return (bool)(!c || (c->flags != tCONS));
}

inline bool IsSymbol(LISP c)
{
  return (bool)(c && (c->flags == tSYM));
}

inline bool IsString(LISP c)
{
  return (bool)(c && (c->flags == tSTR));;
}

inline bool IsInteger(LISP c)
{
  return (bool)(c && (c->flags == tINT));
}

inline bool IsReal(LISP c)
{
  return (bool)(c && (c->flags == tREAL));
}

inline bool IsNumber(LISP c)
{
  return (bool)(c && ((c->flags == tINT) ||
                      (c->flags == tREAL) ));
}

inline bool IsSubr(LISP c)
{
  return (bool)(c && (c->flags == tSUBR));
}

inline bool IsFSubr(LISP c)
{
  return (bool)(c && (c->flags == tFSUBR));
}

inline bool IsExec(LISP c)
{
  return (bool)(c && ((c->flags == tSUBR) ||
                      (c->flags == tFSUBR) ));
}

inline bool IsFile(LISP c)
{
  return (bool)(c && (c->flags == tFILE));
}
//---------------------------------------------------------------------------
// Frames may be save frames or arg frames
// Arg frames contain the evaluated list of args to a (F)SUBR
// Save frames contain the saved environment for each arg to a (F)EXPR
// The saved data looks like this (arg1 being the first formal arg etc.)
//    LISP arg1
//    LISP arg1->value;
//    LISP arg1->prop;
//    LISP arg2
//    LISP arg2->value;
//    LISP arg2->prop;
//    ...
struct Frame
{
  word  last;     // Negated offset to last frame (always +ve)
  char  nargs;    // Number of args in frame
  bool  save;     // Type (Save frame or Arg frame)
  LISP  fun;      // Pointer to function
  LISP  args[];   // Args stored from here
};
typedef Frame* PFrame;

// Globals relating to the stack
extern pointer  stack;     // Base of allocated stack memory
extern pointer  slimit;    // End of allocated stack memory
extern pointer  stop;      // Next free entry
extern PFrame   base;      // Base of current frame

// Prototypes for stack related functions
void Stack(integer size);
void FreeStack();
void NewFrame(LISP fun);
void PopFrame();
void Push(LISP c);
LISP Pop();
void EvList(LISP fargs);
void Bind(LISP formals);
void SweepStack();
void ClearStack();

//---------------------------------------------------------------------------
extern pointer shlimit;    // End of allocated string heap memory
extern pointer shfree;     // Next free string heap entry

// Prototypes for string heap related functions
void   SHeap(integer size);
STRING AllocString(LISP c, int len);
void   SGC();
//---------------------------------------------------------------------------
// Errors always have an error number (-ve if fatal)
// Errors may have a message or can rely upon the handler to provide a
// message.  Explicit messages always override handler messages.
struct Error
{
  int err;
  const char* msg;
};

void LispError(int err, char *msg = NULL);

//---------------------------------------------------------------------------
#endif
