;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; A recursive descent parser for algebraic expressions  ;;
;; AJB  - August 1997                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; This parser is designed to produce results that are
; easily processed:
; Sums and products are collected into a single, 
; multi-operand operation.
; Subtraction and division are treated as the sum of the
; negative and the product of the inverse respectivly 
; Therefore "a+b-c+d" is parsed as
;	(+ a b (- c) d)
; and not as
;	(+ a (+ b (+ (- c) d)))
; 
; The allowed operators are:
; operator operation     location associativity precedence
; +        addition      infix    left          10
; -        subtraction   infix    left          10
; *        mutiplication infix    left          20
; /        division      infix    left          20
; +        null          prefix   right         40
; -        negation      prefix   right         40
; ^        power         infix    right         50
; !        factorial     postfix  left          60
; f()      function      matchfix
; ()       parentheses   matchfix

; Read a token from expr, checking for tokens put back
(defun get-token ((tok))
  (cond ((null tokens) (token expr))
        (t (setq tok (car tokens))
           (setq tokens (cdr tokens))
           tok )))

; Put a token back
(defun put-back (tok)
  (setq tokens (cons tok tokens)) )


; Parse a string expression
(defun parse (s)
  (if (not (stringp s)) (error "Parse expects a string argument"))
  (setq tokens nil)
  (setq expr s)
  (parse-expr) )

; Parse a sum of products
(defun parse-expr ((sum) (rhs) (tok))
  (setq sum (list (parse-product)))
  (loop
    (while (setq tok (get-token)) lhs)
    (until (eql tok ")"))
    (cond
      ((eql tok "+")
       (setq rhs (parse-product))
       (if (and (listp rhs) (eq '+ (car rhs)))
           (setq sum (append sum (cdr rhs)))
           (setq sum (append sum (list rhs))) ))
      ((eql tok "-")
       (setq rhs (parse-product))
       (if (and (listp rhs) (eq '+ (car rhs)))
           (setq sum (append sum (negate (cdr rhs))))
           (setq sum (append sum (list (list '- rhs)))) ))
      (t (error "Bad token")) ))
  (if (cdr sum)
    (cons '+ sum)
    (car sum) ))

; Parse a product of factors
(defun parse-product ((prd) (rhs) (tok))
  (setq prd (list (parse-other)))
  (loop
    (while (setq tok (get-token)))
    (cond
      ((eql tok "*")
       (setq rhs (parse-other))
       (if (and (listp rhs) (eq '* (car rhs)))
           (setq prd (append prd (cdr rhs)))
           (setq prd (append prd (list rhs))) ))
      ((eql tok "/")
       (setq rhs (parse-other))
       (if (and (listp rhs) (eq '* (car rhs)))
           (setq prd (append prd (inverse (cdr rhs))))
           (setq prd (append prd (list (list '1/ rhs)))) ))
      (t (put-back tok) (while nil)) ))
  (if (cdr prd)
    (cons '* prd)
    (car prd) ))

; Parse other operators
(defun parse-other ((lhs) (tok))
  (setq lhs (parse-term))
  (setq tok (get-token))
  ; Factorial?
  (cond
    ((eql tok "!")
     (setq lhs (list 'fact lhs))
     (setq tok (get-token)) ))
  (cond
    ((eql tok "^")
     (list 'power lhs (parse-other)) )
    (t
     (put-back tok)
     lhs )))

; Parse a term
(defun parse-term ((lhs))
  (setq lhs (get-token))
  (cond
    ; A function?
    ((member lhs functions)
     (if (eql (get-token) "(")
         (list lhs (parse-expr))
         (error "function missing argument list") ))
    ; Unary +?
    ((eql lhs "+") (parse-other))
    ; Unary -?
    ((eql lhs "-") (list '- (parse-other)))
    ; A sub expression?
    ((eql lhs "(") (parse-expr))
    ; Just a number or a symbol
    ((or (symbolp lhs) (numberp lhs)) lhs)
    (t (error "parse error"))  ))

; Known functions, extend as desired
(setq functions '(sqrt sin cos tan asin acos atan log ln exp))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Expression output functions

(defun print-expr (e (prec . 0))
  (cond
    ((null e))
    ((atom e)                   (princ e))
    ((eq (car e) '+)            (print-sum (cdr e) prec))      ; prec = 10
    ((eq (car e) '*)            (print-product (cdr e) prec))  ; prec = 20
    ((eq (car e) '1/)           (print-inverse (cdr e) prec))  ; prec = 30
    ((eq (car e) '-)            (print-neg (cdr e) prec))      ; prec = 40
    ((eq (car e) 'power)        (print-power (cdr e) prec))    ; prec = 50
    ((eq (car e) 'fact)         (print-fact (cdr e)))          ; prec = 60
    ((member (car e) functions) (print-func e))                ; prec = infinity
    (t (error "Unknown expression type in print-expr")) ))

(defun print-sum (e (prec . 0))
  (if (> prec 10) (princ lpar)) 
  (print-expr (car e) 10)
  (loop
    (setq e (cdr e))
    (while e)
    (cond 
      ((and (listp (car e)) (eq (caar e) '-))
        (princ '-)
        (print-expr (car (cdar e)) 10) )
      (t
        (princ '+)
        (print-expr (car e) 10) )))
  (if (> prec 10) (princ rpar)) )

(defun print-product (e (prec . 0))
  (if (> prec 20) (princ lpar)) 
  (print-expr (car e) 20)
  (loop
    (setq e (cdr e))
    (while e)
    (cond 
      ((and (listp (car e)) (eq (caar e) '1/))
        (princ '/)
        (print-expr (car (cdar e)) 20) )
      (t
        (princ '*)
        (print-expr (car e) 20) )))
   (if (> prec 20) (princ rpar)) )

 (defun print-inverse(e (prec . 0))
   (if (> prec 30) (princ lpar))
   (princ '1/)
   (print-expr (car e) 30) 
   (if (> prec 30) (princ rpar)) )

 (defun print-neg(e (prec . 0))
   (if (> prec 40) (princ lpar))
   (princ '-)
   (print-expr (car e) 40) 
   (if (> prec 40) (princ rpar)) )

 (defun print-power(e (prec . 0))
   (if (> prec 50) (princ lpar))
   (print-expr (car e) 50)
   (princ '^)
   (print-expr (cadr e) 50) 
   (if (> prec 50) (princ rpar)) )

 (defun print-fact(e (prec . 0))
   (if (> prec 60) (princ lpar))
   (print-expr (car e) 60) 
   (princ '!)
   (if (> prec 60) (princ rpar)) )

 (defun print-func(e (prec . 0))
   (princ (car e))
   (princ lpar)
   (print-expr (cadr e)) 
   (princ rpar) )

     
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Utility functions

; Reciprocal used for division
(defun /1 (x) (/ 1.0 x))

; Negate an atom or list
(defun negate (l)
  (cond
    ((null l) nil)
    ((atom l) (cons '- (list l)))
    (t (cons (negate (car l)) (negate (cdr l)))) ))

; Inverse of an atom or list
(defun inverse (l)
  (cond
    ((null l) nil)
    ((atom l) (cons '/1 (list l)))
    (t (cons (inverse (car l)) (inverse (cdr l)))) ))

(setq lpar '\()
(setq rpar '\))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Test the parser

(defun parser ((e))
(loop
  (princ "expression: ")
  (setq e (readline))
  (while (not (eql e "")) "done")
  (printc "result:     " (parse e)) ))

(setq t1 '( "a+b"
            "a-b"
            "a*b"
            "a/b"
            "a+b-c+d"
            "a*b/c*d"
            "120"
            "120.0"
            "-240.324E-6"
            "12*a+4*b"
            "sin(2*pi)-tan(32*2*pi/180)"
            "(a+b) / (a-b)"
            "exp(ln(x)*3)"
            "a^b^c"
            "-x^2-b*x-c" ))

(defun test (tests (cpy) (e))
  (cond
    ((null tests) nil)
    ( t
      ; force a copy of the string because parse is destructive
      (setq cpy (concat (car tests) "")) 
      (printc " expr: " cpy " parse: " (setq e (parse cpy)))
      (print-expr e)
      (print)
      (test (cdr tests)) )))

; Setup for interactive testing
(setq ft '("+" a ("+" b ("+" c d))))
(backtrace 3)
;(parser)


